import bz2
import gzip
import json
import pytz

from datetime import datetime
from jinja2 import Environment, PackageLoader
from itertools import chain
from os.path import join, dirname

from .utils import compute_checksums
from . import __version__

from conda.models.version import normalized_version


index_keys = (
    "repodata.json",
    "repodata.json.bz2",
    "repodata.json.gz",
    "index.html",
    "index.html.gz",
)
channel_keys = ("style.css", "icon.png", "index.html", "index.html.gz")
keys_to_remove = ("fn", "channel", "binstar", "target-triplet")


def _merged_packages(repodata):
    return sorted(
        chain(
            repodata.get("packages", {}).items(),
            repodata.get("packages.conda", {}).items(),
        ),
        key=lambda x: x[0],
    )


def _ts_to_dt(dt):
    if dt > 253402300799:  # 9999-12-31
        dt //= 1000  # convert milliseconds to seconds; see #1988
    return datetime.utcfromtimestamp(dt).replace(tzinfo=pytz.timezone("UTC"))


def _get_jinja2_environment():
    def _hyphenate(text):
        parts = text.split("-", 1)
        if len(parts) == 1:
            return parts[0]
        elif len(parts[0]) < len(parts[1]):
            return parts[0] + "-<br/>" + parts[1]
        else:
            return parts[0] + "<br/>-" + parts[1]

    def _filter_strftime(dt, dt_format):
        if isinstance(dt, (int, float)):
            dt = _ts_to_dt(dt)
        return dt.strftime(dt_format)

    def _filter_add_href(text, link, **kwargs):
        if link:
            kwargs_list = [f'href="{link}"']
            kwargs_list.append(f'alt="{text}"')
            kwargs_list += [f'{k}="{v}"' for k, v in kwargs.items()]
            return "<a {}>{}</a>".format(" ".join(kwargs_list), text)
        else:
            return text

    def _human_bytes(n):
        if n < 1024:
            return "%d B" % n
        k = n / 1024
        if k < 1024:
            return "%d KB" % round(k)
        m = k / 1024
        if m < 1024:
            return "%.1f MB" % m
        g = m / 1024
        return "%.2f GB" % g

    environment = Environment(loader=PackageLoader("anaconda_mirror", "templates"))
    environment.filters["human_bytes"] = _human_bytes
    environment.filters["strftime"] = _filter_strftime
    environment.filters["add_href"] = _filter_add_href
    environment.filters["hyphenate"] = _hyphenate
    environment.trim_blocks = True
    environment.lstrip_blocks = True
    return environment


def _make_channel_index_html(channel, sdict):
    mtime = 0
    subdirs = set()
    packages = {}
    for sub, repodata in sdict.items():
        for fn, prec in _merged_packages(repodata):
            subdirs.add(sub)
            name = prec["name"]
            version = normalized_version(prec["version"])
            timestamp = prec.get("timestamp", 0)
            vrec = packages.get(name)
            if vrec is None:
                packages[name] = {
                    "name": name,
                    "subdirs": set(sub),
                    "version": version,
                    "timestamp": timestamp,
                }
            else:
                vrec["version"] = max(
                    vrec["version"], normalized_version(prec["version"])
                )
                vrec["subdirs"].add(sub)
                vrec["timestamp"] = max(vrec["timestamp"], timestamp)
            mtime = max(mtime, timestamp)
            subdirs.add(sub)
    packages = [packages[k] for k in sorted(packages)]
    subdirs = (["noarch"] if "noarch" in subdirs else []) + sorted(subdirs - {"noarch"})
    environment = _get_jinja2_environment()
    template = environment.get_template("channeldata-index.html.j2")
    rendered_html = template.render(
        channel=channel,
        subdirs=subdirs,
        packages=packages,
        current_time=_ts_to_dt(mtime),
        mirror_version=__version__,
    )
    return rendered_html


def _make_subdir_index_html(channel, subdir, repodata, extra_paths):
    packages = _merged_packages(repodata)
    mtime = max(v.get("timestamp", 0) for k, v in packages) if packages else 0
    for prec in extra_paths.values():
        prec["timestamp"] = mtime
    environment = _get_jinja2_environment()
    template = environment.get_template("subdir-index.html.j2")
    return template.render(
        channel=channel,
        subdir=subdir,
        packages=packages,
        current_time=_ts_to_dt(mtime),
        extra_paths=extra_paths,
        mirror_version=__version__,
    )


def build_channel_indices(cname, repodatas):
    index_dict = {}
    index_html = _make_channel_index_html(cname, repodatas)
    data_b = index_html.encode("utf-8")
    for fn in channel_keys:
        if fn == "index.html":
            t_data = data_b
        elif fn == "index.html.gz":
            t_data = gzip.compress(data_b, mtime=0)
        else:
            fpath = join(dirname(__file__), "templates", fn)
            with open(fpath, "rb") as fp:
                t_data = fp.read()
        irec = compute_checksums(t_data)
        irec["_data"] = t_data
        index_dict[fn] = irec
    return index_dict


def build_subdir_indices(cname, platform, repodata):
    if platform != "noarch":
        if not repodata["packages"] and not repodata["packages.conda"]:
            return {}
    index_dict = {}
    repodata.pop("extras", None)
    repodata.pop("indices", None)
    for pkey in ("packages", "packages.conda"):
        for prec in repodata[pkey].values():
            for key in keys_to_remove:
                prec.pop(key, None)
    data_str = json.dumps(repodata, sort_keys=True, separators=(",", ":"))
    data_b = data_str.encode("ascii")
    for fn in index_keys:
        if not fn.startswith("repodata"):
            continue
        if fn.endswith(".gz"):
            t_data = gzip.compress(data_b, mtime=0)
        elif fn.endswith(".bz2"):
            t_data = bz2.compress(data_b)
        else:
            t_data = data_b
        irec = compute_checksums(t_data)
        irec["_data"] = t_data
        index_dict[fn] = irec
    index_html = _make_subdir_index_html(cname, platform, repodata, index_dict)
    data_b = index_html.encode("utf-8")
    for fn in index_keys:
        if fn.startswith("repodata"):
            continue
        elif not fn.startswith("index"):
            if platform != "noarch":
                continue
            fpath = join(dirname(__file__), "templates", fn)
            with open(fpath, "rb") as fp:
                t_data = fp.read()
        elif fn.endswith(".gz"):
            t_data = gzip.compress(data_b, mtime=0)
        else:
            t_data = data_b
        irec = compute_checksums(t_data)
        irec["_data"] = t_data
        index_dict[fn] = irec
    return index_dict



import sys

from urllib.parse import urlparse

from .common import _api
from .utils import MirrorException, fp_or_bytes


def get_repo_api(config):
    dest_site = config.dest_site
    if not dest_site:
        raise MirrorException("No Artifactory repository specified.")
    scheme, netloc, path, *remains = urlparse(dest_site)
    if path and not (scheme or netloc or any(remains)):
        if "/" in path:
            netloc, path = path.split("/", 1)
        else:
            netloc, path = path, "/artifactory"
    if not netloc:
        raise MirrorException("Invalid dest_site: %s" % dest_site)
    path = path.strip("/")
    if path in ("", "artifactory"):
        path = "artifactory"
    elif not path.startswith("artifactory/"):
        raise MirrorException("Path must start with /artifactory: %s" % dest_site)
    base = tuple(path.split("/")[1:])
    dest_site = (scheme or "https") + "://" + netloc + "/artifactory"
    config.dest_site = dest_site
    config.api_url = dest_site + "/api"
    return base


def begin_channel(base, cname, dry_run):
    repo = base[0] if base else cname.split("/", 1)[0]
    response = _api("get", "repositories", repo)
    if response["rclass"] != "local":
        raise MirrorException("Must be a local repository: %s" % repo)


def list_objects(base, cname, platform=None, repodata=None):
    # Also grab information a listing of the actual files
    localdata = _api("get", "storage", *base, cname, platform, accept_404=True)
    packages = {}
    repodata = repodata or {}
    s1, s2 = repodata.get("packages", {}), repodata.get("packages.conda", {})
    for frec in localdata.get("children", ()):
        if frec["folder"]:
            continue
        fn = frec["uri"].lstrip("/")
        prec = packages[fn] = {}
        # Let's not waste an API call to get the checksums for packages that
        # are in the index. Instead we just trust the index. If we could do a
        # bulk listing, as we can with other destinations like Nexus or S3,
        # it would be worth the effort. That said, creating the empty dictionary
        # above is important, because it tells our code that the file is there.
        if fn not in s1 and fn not in s2:
            data = _api("get", "storage", *base, cname, platform, fn)
            prec["size"] = int(data["size"])
            prec["md5"] = data["checksums"]["md5"]
            prec["sha1"] = data["checksums"]["sha1"]
            prec["sha256"] = data["checksums"]["sha256"]
    return packages


def get_object(base, *path):
    try:
        return _api("get", "..", *base, *path)
    except MirrorException as exc:
        if " 404 " not in exc.message:
            raise


def del_object(base, *path, metadata=None):
    _api("delete", "..", *base, *path, accept_404=True)


def put_object(base, *path, path_or_data=None, metadata=None):
    headers = {
        "X-Checksum-" + csum: metadata[csum.lower()]
        for csum in ("Sha256", "Sha1", "MD5")
        if metadata and csum.lower() in metadata
    }
    with fp_or_bytes(path_or_data) as data:
        _api(
            "put",
            "..",
            *base,
            *path,
            headers=headers,
            data=data,
        )


def main():
    from .common import main as _main

    _main(sys.modules[__name__])



import sys

from collections import namedtuple
from urllib.parse import urlparse

from .common import _api, _content_data
from .utils import MirrorException, fp_or_bytes

from requests_toolbelt.multipart.encoder import MultipartEncoder


NexusApi = namedtuple("NexusApi", ["base", "components"])


def get_repo_api(config):
    dest_site = config.dest_site
    if not dest_site:
        raise MirrorException("No Nexus repository specified.")
    scheme, netloc, path, *remains = urlparse(dest_site)
    if path and not (scheme or netloc or any(remains)):
        if "/" in path:
            netloc, path = path.split("/", 1)
        else:
            netloc, path = path, ""
    if not netloc:
        raise MirrorException("Invalid destination site: %s" % dest_site)
    path = "/" + path.lstrip("/")
    if path == "/":
        path = "/repository"
    if not path.startswith("/repository"):
        msg = "Invalid destination site: %s" % dest_site
        msg += "\nPath must begin with /repository."
        raise MirrorException(msg)
    base = tuple(path.strip("/").split("/"))[1:]
    base_url = (scheme or "https") + "://" + netloc
    config.api_url = base_url + "/service/rest/v1"
    return NexusApi(base, {})


def _join(api, *args):
    args = tuple(a for a in args if a)
    return "/".join(api.base + args).split("/", 1)


def begin_channel(api, cname, dry_run):
    repo, path = _join(api, cname)
    _api("get", "repositories", "raw", "hosted", repo)
    cnt = ""
    while True:
        data = _api("get", "components?repository=" + repo + cnt)
        for irec in data.get("items", []):
            item = irec["assets"][0]
            group, fn = item["path"].rsplit("/", 1)
            grec = api.components.setdefault(group, {})
            grec[fn] = item["checksum"]
            grec[fn]["id"] = irec["id"]
            grec[fn]["url"] = "/" + item["downloadUrl"].split("/", 3)[-1]
        cnt = data.get("continuationToken")
        if not cnt:
            break
        cnt = "&continuationToken=" + data["continuationToken"]


def list_objects(api, *path, repodata=None):
    repo, path = _join(api, *path)
    return api.components.get(path, {})


def get_object(api, *path):
    fn = path[-1]
    repo, path = _join(api, *path[:-1])
    if fn in api.components.get(path, {}):
        return _api("get", api.components[path][fn]["url"])


def del_object(api, *path, metadata):
    fn = path[-1]
    packages = list_objects(api, *path[:-1])
    if fn in packages:
        _api("delete", "components", packages[fn]["id"])


def put_object(api, *path, path_or_data=None, metadata=None):
    fn = path[-1]
    repo, path = _join(api, *path[:-1])
    content_data = _content_data(fn)
    content_type = content_data.pop("Content-Type")
    with fp_or_bytes(path_or_data) as data:
        mdata = MultipartEncoder(
            fields={
                "raw.asset1": (fn, data, content_type, content_data),
                "raw.directory": (None, path),
                "raw.asset1.filename": (None, fn),
            }
        )
        _api(
            "post",
            "components?repository=" + repo,
            data=mdata,
            headers={"Content-Type": mdata.content_type},
        )


def main():
    from .common import main as _main

    _main(sys.modules[__name__])



print("import: 'anaconda_mirror'")
import anaconda_mirror

print("import: 'anaconda_mirror.api_v4'")
import anaconda_mirror.api_v4

print("import: 'anaconda_mirror.api_v5'")
import anaconda_mirror.api_v5

print("import: 'anaconda_mirror.api_v6'")
import anaconda_mirror.api_v6

print("import: 'anaconda_mirror.api_local'")
import anaconda_mirror.api_local

print("import: 'anaconda_mirror.api_s3'")
import anaconda_mirror.api_s3




import sys
import json
import base64

from collections import namedtuple
from os.path import basename

from .common import _content_data
from .utils import MirrorException, fp_or_bytes


S3Api = namedtuple("S3Api", ["session", "client", "base"])


def _status(response):
    return response.get("ResponseMetadata", {}).get("HTTPStatusCode", 0)


def get_repo_api(config):
    try:
        from botocore import session as b_session
    except ImportError:
        raise MirrorException("Please install the 'botocore' package.")
    session = b_session.get_session()
    client = session.create_client("s3")
    base = config.dest_site + "/" if config.dest_site else ""
    return S3Api(session, client, base)


def _join(api, *args):
    uri = api.base + "/".join(a for a in args if a)
    if uri.startswith("s3://"):
        uri = uri[5:]
    if "/" in uri:
        bucket, key = uri.split("/", 1)
    else:
        bucket, key = uri, ""
    return {"Bucket": bucket, "Key": key}


def _exc(exc):
    return "%s: %s" % (type(exc).__name__, exc)


def list_objects(api, cname, platform=None, repodata=None):
    kwargs = _join(api, cname, platform)
    kwargs = {
        "Bucket": kwargs["Bucket"],
        "MaxKeys": 1000,
        "Prefix": kwargs["Key"] + "/",
        "Delimiter": "/",
    }
    packages = {}
    while True:
        response = api.client.list_objects_v2(**kwargs)
        contents = response.get("Contents", [])
        for frec in contents:
            packages[basename(frec["Key"])] = {
                "size": frec["Size"],
                "md5": frec["ETag"].strip('"'),
            }
        if len(contents) < 1000:
            break
        kwargs["StartAfter"] = contents[-1]["Key"]
    return packages


def get_object(api, *path):
    kwargs = _join(api, *path)
    try:
        response = api.client.get_object(**kwargs)
        return json.loads(response["Body"].read())
    except api.client.exceptions.NoSuchKey:
        pass


def del_object(api, *path, metadata=None):
    kwargs = _join(api, *path)
    response = api.client.delete_object(**kwargs)
    if _status(response) not in (200, 204, 404):
        raise MirrorException(str(response))


def put_object(api, *path, path_or_data=None, metadata=None):
    kwargs = _join(api, *path)
    if metadata:
        kwargs["ContentLength"] = metadata["size"]
        sha256 = base64.b64encode(bytes.fromhex(metadata["sha256"])).decode("ascii")
        kwargs["ChecksumSHA256"] = sha256
    content_data = _content_data(path[-1])
    kwargs.update((k.replace("-", ""), v) for k, v in content_data.items())
    with fp_or_bytes(path_or_data) as fp:
        response = api.client.put_object(Body=fp, **kwargs)
    if _status(response) != 200:
        raise MirrorException(str(response))


def main():
    from .common import main as _main

    _main(sys.modules[__name__])



import json
import sys

from .common import _api
from .conda_interface import find_token, set_repo_token
from .utils import MirrorException, site_to_url


def get_repo_api(config):
    try:
        from repo_cli.utils.config import get_config

        r_config = get_config()
    except ImportError:
        raise MirrorException("Please install the 'conda-repo-cli' package.")
    sites = r_config.get("sites", {})
    ds = config.dest_site
    if not ds and not sites:
        raise MirrorException(
            "No repositories have been connected to conda-repo-cli. "
            "Please configure at least one repo before proceeding."
        )
    if not ds:
        ds = r_config.get("default_site")
    if not ds:
        ds = "default" if "default" in sites else next(iter(sites))
    if ds not in sites:
        # If dest_site is a valid domain name or URL, use that to
        # determine the API url of the site.
        url = site_to_url(ds, "api")
        if not url:
            raise MirrorException(
                "Site not found: %s.\n"
                "Please log in using conda-repo-cli or by supplying a token\n"
                "or username/password combination directly to anaconda-mirror." % ds
            )
        # See if it matches a pre-configured entry
        for k, v in sites.items():
            if v.get("url") == url:
                ds = k
                break
    if ds in sites:
        config.dest_site = ds
        url = sites[ds]["url"]
        if config.verify_ssl_dst and "ssl_verify" in sites[ds]:
            config.verify_ssl_dst = sites[ds]["ssl_verify"]
    if config.username and config.password:
        try:
            from repo_cli.utils.api import RepoApi
        except ImportError:
            raise MirrorException("Please install the 'conda-repo-cli' package.")
        t_api = RepoApi(url, verify_ssl=config.verify_ssl)
        t_api.login(config.username, config.password)
        config.password = t_api.get_user_token(force_scopes=None)
        config.username = None
    if config.password:
        set_repo_token(url, config.password)
        config.password = None
    repo_token = find_token(url)
    if not repo_token:
        url = "/".join(url.rsplit("/", 3)[:3])
        raise MirrorException(
            "Cannot authenticate to %s.\n"
            "Please log in using conda-repo-cli or by supplying a token\n"
            "or username/password combination directly to anaconda-mirror." % url
        )
    config.api_url = url
    config.dest_channel_alias = url + "/repo"
    config.api_headers = {"X-Auth": repo_token}
    result = {"auto_freeze": getattr(config, "auto_freeze", True)}
    return result


def _set_frozen(api, cname, frozen):
    if api["frozen"] != frozen and api["auto_freeze"]:
        data = {"indexing_behavior": "frozen" if frozen else "default"}
        _api("put", "channels", cname, json=data, throw=False)
        api["frozen"] = frozen


def begin_channel(api, cname, dry_run=False):
    response = _api("get", "channels", cname)
    fmode = response.get("indexing_behavior")
    if fmode is None:
        api["auto_freeze"] = False
    api["frozen"] = fmode == "frozen"
    if dry_run:
        return
    # Trigger an index generation
    _set_frozen(api, cname, False)
    if not api["frozen"]:
        _api("put", "channels", cname, "artifacts", "conda")


def patch_package(api, cname, platform, fn, metadata):
    _set_frozen(api, cname, True)
    rec = metadata.copy()
    rec["fn"] = fn
    data = {"repodata_record": json.dumps(rec)}
    _api("patch", "repo", cname, platform, fn, json=data)


def finish_channel(api, cname, dry_run):
    if dry_run:
        return
    _set_frozen(api, cname, False)
    # Trigger an index generation
    if not api["frozen"]:
        _api("put", "channels", cname, "artifacts", "conda")


def del_object(api, cname, platform, fn, metadata=None):
    _set_frozen(api, cname, False)
    record = {
        "action": "delete",
        "items": [
            {
                "name": metadata["name"],
                "family": "conda",
                "ckey": "%s/%s" % (platform, fn),
            }
        ],
    }
    _api("put", "channels", cname, "artifacts", "bulk", json=record, accept_404=True)


def put_object(api, cname, platform, fn, path_or_data=None, metadata=None):
    _set_frozen(api, cname, True)
    with open(path_or_data, "rb") as fp:
        multipart_form_data = {
            "content": (fn, fp),
            "filetype": (None, "conda1"),
            "size": (None, metadata["size"]),
            "repodata_record.json": (None, json.dumps(metadata)),
        }
        _api("post", "channels", cname, "artifacts", files=multipart_form_data)


def main():
    from .common import main as _main

    _main(sys.modules[__name__])
